----------------------------------------
-- Motion steps FIFO for Mach3 stuff. --
----------------------------------------

--------------------------------------------------------------------------
--			LICENSE and Terms of Use.			--
--									--
-- First of all, this is NOT GPL or Free Software. It is open so you	--
-- have the VHDL source code. There is also the source code for the	--
-- Windows Engine for Mach3 CNC software available from KSI Labs, LLC	--
-- website, http://www.ksilabs.com.					--
--									--
-- You are free to use this particular VHDL code, make modifications to	--
-- customize it for your own needs, fix inevitable bugs, port it to 	--
-- other CNC software and do anything else as long as your derivative	--
-- work that is based on this code or uses parts or pieces thereof is	--
-- made to be used with any KSI Labs, LLC products.			--
--									--
-- We do NOT grant you a permission to use any parts of this code in	--
-- any products that are designed for products not manufactured or	--
-- distributed by KSI Labs, LLC. We might give such a permission on	--
-- case-by-case basis but it is not legal to use this code or bits and	--
-- pieces thereof with products not coming from KSI Labs, LLC without	--
-- our explicit written permission.					--
--									--
-- The original copyright notices must be included in all derivative	--
-- works. You do NOT have to ask our permission to make any changes to	--
-- this code and you are not under obligation to give us any of your	--
-- work back as long as you are using it for your own needs.		--
--									--
-- However we do NOT give our permission to distribute your derivative	--
-- work unless all original copyrights are retained, all the conditions	--
-- of this license are met (use only with KSI Labs, LLC products) and	--
-- full source code for such derivative work is available for free to	--
-- everybody without any further conditions including but not limited	--
-- to signing any kinds of NDA. We explicitely deny any permission to	--
-- distribute any derivative work based on this code with closed source	--
-- binary-only fragments or include bits or pieces of this code in such	--
-- products. That essentially means that it is NOT legal to use it with --
-- ANY proprietary HDL megafunctions, megacores, etc. if the resulting	--
-- product is distributed. It IS legal to use whatever you want if the	--
-- resulting product is made for use with a KSI Labs, LLC product AND	--
-- that resulting product is not distributed. Exclusions can be granted --
-- by KSI Labs, LLC on case-by-case basis.				--
--------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.mach3_data_types.all;

entity mach3_fifo is
generic (
	fifo_depth	: natural := 20
);
port (
	fifo_in		: in std_logic_vector(11 downto 0);
	fifo_out	: out std_logic_vector(11 downto 0);
	fifo_push	: in std_logic;
	fifo_pop	: in std_logic;
	fifo_purge	: in std_logic;
	fifo_full	: out std_logic;
	fifo_empty	: out std_logic;
	fifo_ovf_irq	: out std_logic;
	fifo_ovf_clr	: in std_logic;
	fifo_room	: out std_logic_vector(4 downto 0);
	clk		: in std_logic;
	rst		: in std_logic
);
end mach3_fifo;

architecture guts of mach3_fifo is

	-------------
	-- Signals --
	-------------

	type fifo_data is array (fifo_depth-1 downto 0) of std_logic_vector(11 downto 0);
	signal fifo		: fifo_data;
	
	signal wr_ptr		: natural range 0 to fifo_depth-1;
	signal f_full		: std_logic;
	signal f_empty		: std_logic;
	signal f_ovf		: std_logic;
	signal do_pop		: std_logic;

begin

	fifo_proc : process (rst, clk)
	begin
		if (rst='1') then
			wr_ptr <= fifo_depth-1;
			f_full <= '0';
			f_ovf <= '0';
			do_pop <= '0';
			fifo(fifo_depth-1) <= (others => '0');
		elsif (rising_edge(clk)) then
			if (fifo_purge='1') then
				wr_ptr <= fifo_depth-1;
				f_full <= '0';
				f_ovf <= '0';
				do_pop <= '0';
				fifo(fifo_depth-1) <= (others => '0');
			elsif (fifo_push='1') then
				if (wr_ptr/=0) then
					fifo(wr_ptr) <= fifo_in;
					wr_ptr <= wr_ptr - 1;
				elsif (f_full/='1') then
					fifo(wr_ptr) <= fifo_in;
					f_full <= '1';
				else
					f_ovf <= '1';
				end if;
				do_pop <= fifo_pop;
			elsif (fifo_ovf_clr='1') then
				f_ovf <= '0';
				do_pop <= fifo_pop;
			elsif ((fifo_pop='1' or do_pop='1') and f_empty='0') then
				if (wr_ptr/=fifo_depth-2) then
				-- No shift on last entry; keep it
					for i in fifo_depth-1 downto 1 loop
						fifo(i) <= fifo(i-1);
					end loop;
				end if;
				if (f_full='1') then
				-- No pointer increment if FIFO was full
					f_full <= '0';
				else
					wr_ptr <= wr_ptr + 1;
				end if;
				f_ovf <= '0';
				do_pop <= '0';
			end if;
		end if;
	end process;

	f_empty <= '1' when (wr_ptr=fifo_depth-1) else '0';
	fifo_out <= fifo(fifo_depth-1);
	fifo_full <= f_full;
	fifo_empty <= f_empty;
	fifo_ovf_irq <= f_ovf;
	fifo_room <= std_logic_vector(to_unsigned(wr_ptr+1,5))
		when (f_full='0') else (others => '0');
end guts;
