------------------------------------------------------------------------
-- ADC SPI interface. Ugly and hardcoded but easy on the host system  --
-- and fully standalone, no programming required. Made for a specific --
-- chip, TI TLV1544 that is installed on CENTIPEDE-BRK board and      --
-- hooked up to LSS21 on SPI2 interface.                              --
--                                                                    --
-- Monstrous 2-level FSM - goes through cycles (actions) such as INIT --
-- or LoadADC1AddressReadADC0Data etc.; each of cycles in turn goes   --
-- through the same series of states implementing TLV1544-specific    --
-- SPI interface. Cycles (actions) are outer loop, states (SPI) are   --
-- inner loop running inside each cycle. SPI transactions are only    --
-- 10 bit long, there is no need for more.                            --
------------------------------------------------------------------------

--------------------------------------------------------------------------
--			LICENSE and Terms of Use.			--
--									--
-- First of all, this is NOT GPL or Free Software. It is open so you	--
-- have the VHDL source code. There is also the source code for the	--
-- Windows Engine for Mach3 CNC software available from KSI Labs, LLC	--
-- website, http://www.ksilabs.com.					--
--									--
-- You are free to use this particular VHDL code, make modifications to	--
-- customize it for your own needs, fix inevitable bugs, port it to 	--
-- other CNC software and do anything else as long as your derivative	--
-- work that is based on this code or uses parts or pieces thereof is	--
-- made to be used with any KSI Labs, LLC products.			--
--									--
-- We do NOT grant you a permission to use any parts of this code in	--
-- any products that are designed for products not manufactured or	--
-- distributed by KSI Labs, LLC. We might give such a permission on	--
-- case-by-case basis but it is not legal to use this code or bits and	--
-- pieces thereof with products not coming from KSI Labs, LLC without	--
-- our explicit written permission.					--
--									--
-- The original copyright notices must be included in all derivative	--
-- works. You do NOT have to ask our permission to make any changes to	--
-- this code and you are not under obligation to give us any of your	--
-- work back as long as you are using it for your own needs.		--
--									--
-- However we do NOT give our permission to distribute your derivative	--
-- work unless all original copyrights are retained, all the conditions	--
-- of this license are met (use only with KSI Labs, LLC products) and	--
-- full source code for such derivative work is available for free to	--
-- everybody without any further conditions including but not limited	--
-- to signing any kinds of NDA. We explicitely deny any permission to	--
-- distribute any derivative work based on this code with closed source	--
-- binary-only fragments or include bits or pieces of this code in such	--
-- products. That essentially means that it is NOT legal to use it with --
-- ANY proprietary HDL megafunctions, megacores, etc. if the resulting	--
-- product is distributed. It IS legal to use whatever you want if the	--
-- resulting product is made for use with a KSI Labs, LLC product AND	--
-- that resulting product is not distributed. Exclusions can be granted --
-- by KSI Labs, LLC on case-by-case basis.				--
--------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity tlv1544_adc is
port (
	spi_ok		: out std_logic;
	ch0_done	: out std_logic;
	ch1_done	: out std_logic;
	ch2_done	: out std_logic;
	ch3_done	: out std_logic;
	ch0_data	: out std_logic_vector(9 downto 0);
	ch1_data	: out std_logic_vector(9 downto 0);
	ch2_data	: out std_logic_vector(9 downto 0);
	ch3_data	: out std_logic_vector(9 downto 0);
	ss_pin		: out std_logic;
	sclk_pin	: out std_logic;
	mosi_pin	: out std_logic;
	miso_pin	: in std_logic;
	clk		: in std_logic;
	nrst		: in std_logic
);
end tlv1544_adc;

architecture guts of tlv1544_adc is

	-------------
	-- Signals --
	-------------

	signal set_spi_ok	: std_logic;
	signal spi_ok_sig	: std_logic;
	signal chk_buf		: std_logic_vector(9 downto 0);
	signal cmd		: std_logic_vector(3 downto 0);
	signal bit_cntr		: unsigned(3 downto 0);
	signal do_ld_thr	: std_logic;
	signal do_shift		: std_logic;
	signal do_rd_bit	: std_logic;
	signal ch0_reg_ld	: std_logic;
	signal ch1_reg_ld	: std_logic;
	signal ch2_reg_ld	: std_logic;
	signal ch3_reg_ld	: std_logic;
	signal chkbuf_ld	: std_logic;
	signal data_in		: std_logic;
	signal do_dly		: std_logic;
	signal cy_end		: std_logic;
	signal dly_cntr		: unsigned(3 downto 0);
	signal thr		: std_logic_vector(9 downto 0);

	type fsm_state is (IDLE, START, LOAD, LATCH, SHIFT, DELAY);
	type fsm_cycle is (INIT, DUMMY, CHECK200, CHECK_IF_OK,
				L0_R3, L1_R0, L2_R1, L3_R2);
			
	signal st_cur	: fsm_state;
	signal st_nxt	: fsm_state;
	signal cy_cur	: fsm_cycle;
	signal cy_nxt	: fsm_cycle;
begin

	--------------------------------------
	-- Top (sequential) part of ADC FSM --
	--------------------------------------
	fsm_top : process(nrst, clk)
		variable rhr	: std_logic_vector(9 downto 0);
	begin
		if (nrst='0') then
			cy_cur <= INIT;
			st_cur <= IDLE;
			spi_ok_sig <= '0';
			ch0_done <= '0';
			ch1_done <= '0';
			ch2_done <= '0';
			ch3_done <= '0';
			rhr := (others => '0');
			thr <= (others => '0');
			bit_cntr <= (others => '0');
			dly_cntr <= (others => '0');
		elsif (rising_edge(clk)) then
			if (do_dly='1') then
				dly_cntr <= dly_cntr + 1;
			else
				if (do_ld_thr='1') then
					thr <= cmd & "000000";
					rhr := (others => '0');
					bit_cntr <= (others => '0');
					if (ch0_reg_ld='1') then
						ch0_done <= '0';
					elsif (ch1_reg_ld='1') then
						ch1_done <= '0';
					elsif (ch2_reg_ld='1') then
						ch2_done <= '0';
					elsif (ch3_reg_ld='1') then
						ch3_done <= '0';
					end if;
				elsif (do_rd_bit='1') then
					rhr(0) := miso_pin;
					if (do_shift='1') then
						thr <= thr(8 downto 0) & '0';
						rhr := rhr(8 downto 0) & '0';
						bit_cntr <= bit_cntr + 1;
					end if;
				end if;

				if (data_in='1') then
					if (ch0_reg_ld='1') then
						ch0_data <= rhr;
						if (spi_ok_sig='1') then
							ch0_done <= '1';
						end if;
					elsif (ch1_reg_ld='1') then
						ch1_data <= rhr;
						if (spi_ok_sig='1') then
							ch1_done <= '1';
						end if;
					elsif (ch2_reg_ld='1') then
						ch2_data <= rhr;
						if (spi_ok_sig='1') then
							ch2_done <= '1';
						end if;
					elsif (ch3_reg_ld='1') then
						ch3_data <= rhr;
						if (spi_ok_sig='1') then
							ch3_done <= '1';
						end if;
					elsif (chkbuf_ld='1') then
						chk_buf <= rhr;
					end if;
				end if;

				if (cy_end='1') then
					if (chkbuf_ld='1') then
						spi_ok_sig <= set_spi_ok;
					end if;
					cy_cur <= cy_nxt;
					dly_cntr <= (others => '0');
					st_cur <= IDLE;
				else
					st_cur <= st_nxt;
				end if;
			end if;
		end if;
	end process;


	------------------------------------------------------
	-- SPI FSM (inner loop) bottom half (combinatorial) --
	------------------------------------------------------
	fsm_st_bh : process(st_cur, bit_cntr, dly_cntr)
	begin
		cy_end <= '0';
		do_ld_thr <= '0';
		do_shift <= '0';
		do_rd_bit <= '0';
		data_in <= '0';
		do_dly <= '0';
		case st_cur is
			when IDLE =>
				st_nxt <= START;
			when START =>
				do_ld_thr <= '1';
				st_nxt <= LOAD;
			when LOAD =>
				st_nxt <= LATCH;
			when LATCH =>
				do_rd_bit <= '1';
				if (bit_cntr=x"9") then
					st_nxt <= DELAY;
					data_in <= '1';
				else
					do_shift <= '1';
					st_nxt <= SHIFT;
				end if;
			when SHIFT =>
				st_nxt <= LATCH;
			when DELAY =>
				if (dly_cntr(3)='1') then
					cy_end <= '1';
					do_dly <= '0';
					st_nxt <= IDLE;
				else
					do_dly <= '1';
					st_nxt <= DELAY;
				end if;
		end case;
	end process;


	---------------------------------------------------------
	-- Action FSM (outer loop) bottom half (combinatorial) --
	---------------------------------------------------------
	fsm_cy_bh : process(cy_cur, chk_buf)
	begin
		ch0_reg_ld <= '0';
		ch1_reg_ld <= '0';
		ch2_reg_ld <= '0';
		ch3_reg_ld <= '0';
		chkbuf_ld <= '0';
		set_spi_ok <= '0';
		case cy_cur is
			when INIT =>
				cmd <= x"9";	-- Set fast rate
				cy_nxt <= DUMMY;
			when DUMMY =>
				cmd <= x"0";
				cy_nxt <= CHECK200;
			when CHECK200 =>
				cmd <= x"b";
				cy_nxt <= CHECK_IF_OK;
			when CHECK_IF_OK =>
				cmd <= x"0";
				chkbuf_ld <= '1';
				-- Ignore 2 LSBits to avoid false positives
				if (chk_buf(9 downto 2)="10000000" or
					chk_buf(9 downto 2)="01111111")
				then
					set_spi_ok <= '1';
					cy_nxt <= L1_R0;
				else
					cy_nxt <= INIT;
				end if;
			when L1_R0 =>
				cmd <= x"2";
				ch0_reg_ld <= '1';
				cy_nxt <= L2_R1;
			when L2_R1 =>
				cmd <= x"4";
				ch1_reg_ld <= '1';
				cy_nxt <= L3_R2;
			when L3_R2 =>
				cmd <= x"6";
				ch2_reg_ld <= '1';
				cy_nxt <= L0_R3;
			when L0_R3 =>
				cmd <= x"b";
				ch3_reg_ld <= '1';
				cy_nxt <= CHECK_IF_OK;
		end case;
	end process;

	ss_pin <= '1' when (st_cur=IDLE or st_cur=DELAY) else '0';
	sclk_pin <= '1' when (st_cur=LATCH) else '0';
	mosi_pin <= thr(9);
	spi_ok <= spi_ok_sig;
end guts;
